"""Test that the output_blocks settings work."""
from ou_container_builder.settings import OutputBlocks, OutputBlock


def test_output_block_conversion() -> None:
    settings = OutputBlocks(build=["RUN echo 'Test'"])
    assert settings.build == [OutputBlock(block="RUN echo 'Test'")]



"""Test that the builder works."""
import os
import subprocess

from typer import Typer
from typer.testing import CliRunner


def test_build_code_server(runner: CliRunner, app: Typer):
    """Test that building the code-server demo works."""
    cwd = os.getcwd()
    try:
        os.chdir(os.path.join("tests", "demos", "code_server"))
        result = runner.invoke(app, ["clean"])
        assert not os.path.exists("Dockerfile")
        assert not os.path.exists("ou-builder-build")
        result = runner.invoke(app, ["build", "--tag", "code_server:test"])
        assert result.exit_code == 0
        assert not os.path.exists("Dockerfile")
        assert not os.path.exists("ou-builder-build")
    finally:
        os.chdir(cwd)
        subprocess.run(["buildah", "rmi", "code_server:test"])


def test_build_jupyterlab_v3(runner: CliRunner, app: Typer):
    """Test that building the JupyterLab v3 demo works."""
    cwd = os.getcwd()
    try:
        os.chdir(os.path.join("tests", "demos", "jupyterlab", "v3"))
        result = runner.invoke(app, ["clean"])
        assert not os.path.exists("Dockerfile")
        assert not os.path.exists("ou-builder-build")
        result = runner.invoke(app, ["build", "--tag", "jupyterlab_v3:test"])
        assert result.exit_code == 0
        assert not os.path.exists("Dockerfile")
        assert not os.path.exists("ou-builder-build")
    finally:
        os.chdir(cwd)
        subprocess.run(["buildah", "rmi", "jupyterlab_v3:test"])


def test_build_jupyterlab_v4(runner: CliRunner, app: Typer):
    """Test that building the JupyterLab v4 demo works."""
    cwd = os.getcwd()
    try:
        os.chdir(os.path.join("tests", "demos", "jupyterlab", "v4"))
        result = runner.invoke(app, ["clean"])
        assert not os.path.exists("Dockerfile")
        assert not os.path.exists("ou-builder-build")
        result = runner.invoke(app, ["build", "--tag", "jupyterlab_v4:test"])
        assert result.exit_code == 0
        assert not os.path.exists("Dockerfile")
        assert not os.path.exists("ou-builder-build")
    finally:
        os.chdir(cwd)
        subprocess.run(["buildah", "rmi", "jupyterlab_v4:test"])


def test_build_custom_apt_key_dearmor(runner: CliRunner, app: Typer):
    """Test that building the custom_apt_key_dearmor demo works."""
    cwd = os.getcwd()
    try:
        os.chdir(os.path.join("tests", "demos", "custom_apt_key_dearmor"))
        result = runner.invoke(app, ["clean"])
        assert not os.path.exists("Dockerfile")
        assert not os.path.exists("ou-builder-build")
        result = runner.invoke(app, ["build", "--tag", "custom_apt_key_dearmor:test"])
        assert result.exit_code == 0
        assert not os.path.exists("Dockerfile")
        assert not os.path.exists("ou-builder-build")
    finally:
        os.chdir(cwd)
        subprocess.run(["buildah", "rmi", "custom_apt_key_dearmor:test"])


def test_build_openrefine(runner: CliRunner, app: Typer):
    """Test that building the openrefine demo works."""
    cwd = os.getcwd()
    try:
        os.chdir(os.path.join("tests", "demos", "openrefine"))
        result = runner.invoke(app, ["clean"])
        assert not os.path.exists("Dockerfile")
        assert not os.path.exists("ou-builder-build")
        result = runner.invoke(app, ["build", "--tag", "openrefine:test"])
        assert result.exit_code == 0
        assert not os.path.exists("Dockerfile")
        assert not os.path.exists("ou-builder-build")
    finally:
        os.chdir(cwd)
        subprocess.run(["buildah", "rmi", "openrefine:test"])



"""Fixtures for the CLI tests."""
from pytest import fixture
from typer import Typer
from typer.testing import CliRunner

from ou_container_builder.__main__ import app as cli_app

collect_ignore = ["test_build.py"]


@fixture()
def runner() -> CliRunner:
    """Provide a CliRunner as a fixture."""
    return CliRunner()


@fixture
def app() -> Typer:
    """Provide the Typer application as a fixture."""
    return cli_app



"""Test the utility functions."""
from ou_container_builder.util import docker_copy_cmd, docker_run_cmd


def test_basic_copy_command() -> None:
    """Test the basic copy command."""
    assert 'COPY ["source", "target"]' == docker_copy_cmd("source", "target")


def test_from_stage_copy() -> None:
    """Test that setting the from_stage parameter works."""
    assert 'COPY --from=build-stage ["source", "target"]' == docker_copy_cmd(
        "source", "target", from_stage="build-stage"
    )


def test_chmod_copy() -> None:
    """Test that chmod of the copy command works."""
    assert 'COPY --chmod=0664 ["source", "target"]' == docker_copy_cmd("source", "target", chmod="0664")


def test_from_stage_chmod_copy() -> None:
    """Test that from_stage and chmod together works."""
    assert 'COPY --from=build-stage --chmod=0664 ["source", "target"]' == docker_copy_cmd(
        "source", "target", from_stage="build-stage", chmod="0664"
    )


def test_run_single_command() -> None:
    """Test that running a single command works."""
    assert "RUN /usr/bin/sleep 60" == docker_run_cmd(["/usr/bin/sleep 60"])


def test_run_multiple_commands() -> None:
    """Test that running multiple commands works."""
    assert "RUN /usr/bin/sleep 60 && \\\n    echo $HOME" == docker_run_cmd(["/usr/bin/sleep 60", "echo $HOME"])



"""Extension packs."""
from importlib import import_module
from importlib.metadata import entry_points
from rich.progress import Progress
from typing import Callable, Any

from ou_container_builder.state import State


EXTENSION_PACKS = {}
for entry_point in entry_points().select(group="ou_container_builder"):
    EXTENSION_PACKS[entry_point.name] = import_module(entry_point.value)


def make_listener(key: str, pack: Any) -> Callable[[State], None]:
    """Create a custom listener for the key and pack."""

    def listener(listener_state: State) -> None:
        if key in listener_state["packs"] and listener_state["packs"][key] is not None:
            pack.init(listener_state)

    return listener


def init(state: State, progress: Progress) -> None:
    """Initialise all extension packs."""
    extension_task = progress.add_task("Initialising packs...", total=len(EXTENSION_PACKS))
    for key, pack in EXTENSION_PACKS.items():
        state.add_listener(f"packs.{key}", make_listener(key, pack))
        progress.update(extension_task, advance=1)
    progress.update(extension_task, visible=False)


def generate(state: State, progress: Progress) -> None:
    """Generate all files and Dockerfile snippets."""
    extension_task = progress.add_task("Generating packs...", total=len(EXTENSION_PACKS))
    for key, pack in EXTENSION_PACKS.items():
        if key in state["packs"] and state["packs"][key] is not None:
            pack.generate(state, progress)
        progress.update(extension_task, advance=1)
    progress.update(extension_task, visible=False)



"""Test the packages settings validation."""
from ou_container_builder.settings import Packages, PipPackageLists, PipPackageEntry


def test_valid_packages_settings():
    """Test that a valid packages configuration passes."""
    settings = Packages(apt=["curl"], pip=["jupyterlab"])
    assert settings.apt.build == ["curl"]
    assert settings.apt.deploy == ["curl"]
    assert settings.pip.system[0].name == "jupyterlab"
    assert settings.pip.system[0].type == "package"
    assert settings.pip.user[0].name == "jupyterlab"
    assert settings.pip.user[0].type == "package"


def test_default_sources_settings():
    """Test that the default sources settings are correct."""
    settings = Packages()
    assert settings.apt.build == []
    assert settings.apt.deploy == []
    assert settings.pip.system == []
    assert settings.pip.user == []


def test_simple_pip_package_conversion() -> None:
    """Test that the Pip package entry conversion works."""
    settings = PipPackageLists(system=["jupyterlab"])
    assert settings.system == [PipPackageEntry(name="jupyterlab", type="package")]



"""Test the environment settings validation."""
from pydantic import ValidationError
from pytest import raises

from ou_container_builder.settings import EnvironmentVariable

from .util import error_locations


def test_valid_environment_settings():
    """Test that a valid environment configuration passes."""
    EnvironmentVariable(name="EXTRA", value="yes")


def test_empty_name_fails():
    """Test that an environment variable with an empty name fails."""
    with raises(ValidationError) as e_info:
        EnvironmentVariable(name="", value="")
    assert ("name",) in error_locations(e_info.value)


def test_null_value_fails():
    """Test that an environment variable with a None value fails."""
    with raises(ValidationError) as e_info:
        EnvironmentVariable(name="TEST", value=None)
    assert ("value",) in error_locations(e_info.value)



"""Test the image settings validation."""
from ou_container_builder.settings import Image


def test_valid_image_settings():
    """Test that a valid module configuration passes."""
    settings = Image(base="debian:stable", user="jovyan")
    assert settings.base == "debian:stable"
    assert settings.user == "jovyan"


def test_default_settings():
    """Test that the default settings are set correctly."""
    settings = Image()
    assert settings.base == "python:3.11-bookworm"
    assert settings.user == "ou"



"""Functionality for installing python packages.

The packages are downloaded and built in the build stage and then the wheels are copied over into the
deploy stage. The system packages are then installed into a venv at /var/lib/ou/python/system.

Packages are installed from requirements files. If packages are provided directly, then a requirements.txt
is generated for those.
"""
import os

from rich.progress import Progress

from ou_container_builder.state import State
from ou_container_builder.util import docker_run_cmd, docker_copy_cmd


STARTUP_INSTALL_SCRIPT = """#/bin/bash
pip install --user --no-index /usr/share/ou/python/user/*.whl
"""


def init(state: State) -> None:
    """Unused."""
    pass


def generate(state: State, progress: Progress) -> None:
    """Generate the python wheel build and install blocks."""
    # If there are any installed packages copy them from the build to the deploy stage
    if len(state["packages"]["pip"]["system"]) > 0 or len(state["packages"]["pip"]["user"]) > 0:
        state.update(
            {
                "output_blocks": {
                    "deploy": [
                        {
                            "block": docker_copy_cmd("/usr/share/ou/python", "/usr/share/ou/python", from_stage="base"),
                            "weight": 161,
                        },
                    ]
                }
            }
        )
    # Install any system python packages
    if len(state["packages"]["pip"]["system"]) > 0:
        system_packages = "\\n".join(
            [pkg["name"] for pkg in state["packages"]["pip"]["system"] if pkg["type"] == "package"]
        )
        requirement_files = " ".join(
            [
                f'-r /usr/share/ou/python/system/requirements-{idx+1}.txt'
                for idx, pkg in enumerate(state["packages"]["pip"]["system"])
                if pkg["type"] == "requirements.txt"
            ]
        )
        state.update(
            {
                "output_blocks": {
                    "build": [
                        {
                            "block": docker_copy_cmd(
                                pkg["name"], f"/usr/share/ou/python/system/requirements-{idx + 1}.txt"
                            ),
                            "weight": 161,
                        }
                        for idx, pkg in enumerate(state["packages"]["pip"]["system"])
                        if pkg["type"] == "requirements.txt"
                    ]
                }
            }
        )
        state.update(
            {
                "output_blocks": {
                    "build": [
                        {
                            "block": docker_run_cmd(
                                [
                                    "mkdir -p /usr/share/ou/python/system",
                                    f'echo "{system_packages}" > /usr/share/ou/python/system/requirements-0.txt',
                                ]
                            ),
                            "weight": 161,
                        },
                        {
                            "block": docker_run_cmd(
                                [
                                    f"pip wheel --no-cache-dir -w /usr/share/ou/python/system -r /usr/share/ou/python/system/requirements-0.txt {requirement_files}",
                                ]
                            ),
                            "weight": 162,
                        },
                    ],
                    "deploy": [
                        {
                            "block": docker_run_cmd(
                                [
                                    "mkdir -p /var/lib/ou/python",
                                    "python -m venv /var/lib/ou/python/system",
                                    "/var/lib/ou/python/system/bin/pip install --no-index --no-deps /usr/share/ou/python/system/*.whl",
                                    "/var/lib/ou/python/system/bin/pip uninstall -y ipykernel",
                                ]
                            ),
                            "weight": 162,
                        },
                    ],
                }
            }
        )
    # Build any user packages
    if len(state["packages"]["pip"]["user"]) > 0:
        user_packages = "\\n".join(
            [pkg["name"] for pkg in state["packages"]["pip"]["user"] if pkg["type"] == "package"]
        )
        # Build and copy the requirements files
        requirement_files = " ".join(
            [
                f'-r /usr/share/ou/python/user/requirements-{idx+1}.txt'
                for idx, pkg in enumerate(state["packages"]["pip"]["user"])
                if pkg["type"] == "requirements.txt"
            ]
        )
        state.update(
            {
                "output_blocks": {
                    "build": [
                        {
                            "block": docker_copy_cmd(
                                pkg["name"], f"/usr/share/ou/python/user/requirements-{idx + 1}.txt"
                            ),
                            "weight": 163,
                        }
                        for idx, pkg in enumerate(state["packages"]["pip"]["user"])
                        if pkg["type"] == "requirements.txt"
                    ]
                }
            }
        )
        # Build the user packages
        state.update(
            {
                "output_blocks": {
                    "build": [
                        {
                            "block": docker_run_cmd(
                                [
                                    'mkdir -p /usr/share/ou/python/user',
                                    f'echo "{user_packages}" > /usr/share/ou/python/user/requirements-0.txt',
                                ]
                            ),
                            "weight": 161,
                        },
                        {
                            "block": docker_run_cmd(
                                [
                                    f"pip wheel --no-cache-dir -w /usr/share/ou/python/user -r /usr/share/ou/python/user/requirements-0.txt {requirement_files}"
                                ]
                            ),
                            "weight": 164,
                        },
                    ]
                }
            }
        )
        # Add the startup install script
        os.makedirs(os.path.join("ou-builder-build", "python"), exist_ok=True)
        with open(os.path.join("ou-builder-build", "python", "010-installing-python-packages"), "w") as out_f:
            out_f.write(STARTUP_INSTALL_SCRIPT)
        state.update(
            {
                "output_blocks": {
                    "deploy": [
                        {
                            "block": docker_copy_cmd(
                                os.path.join("ou-builder-build", "python", "010-installing-python-packages"),
                                "/usr/share/ou/startup.d/010-installing-python-packages",
                            ),
                            "weight": 191,
                        }
                    ]
                }
            }
        )



"""Core builder functionality."""
from rich.progress import Progress

from ou_container_builder.state import State
from ou_container_builder.core import (
    apt,
    args,
    base,
    content,
    environment,
    python,
    rust,
    scripts,
    server,
    services,
    sources,
    user,
    webapp,
)


CORE_PACKS = [apt, args, base, content, environment, python, rust, scripts, server, services, sources, user, webapp]


def init(state: State, progress: Progress) -> None:
    """Initialise all core packs."""
    core_task = progress.add_task("Initialising core...", total=len(CORE_PACKS))
    for pack in CORE_PACKS:
        pack.init(state)
        progress.update(core_task, advance=1)
    progress.update(core_task, visible=False)


def generate(state: State, progress: Progress) -> None:
    """Generate all files and Dockerfile snippets."""
    core_task = progress.add_task("Generating core...", total=len(CORE_PACKS))
    for pack in CORE_PACKS:
        pack.generate(state, progress)
        progress.update(core_task, advance=1)
    progress.update(core_task, visible=False)



"""Functionality for setting up the user in both stages."""
from rich.progress import Progress

from ou_container_builder.state import State
from ou_container_builder.util import docker_run_cmd


def init(state: State) -> None:
    """Unused."""
    pass


def generate(state: State, progress: Progress) -> None:
    """Generate the user setup for the two stages."""
    state.update(
        {
            "output_blocks": {
                "build": [
                    {
                        "block": docker_run_cmd(
                            ["mkdir /home/$USER && useradd -u $UID -g $GID -d $HOME -m -s /bin/bash $USER"]
                        ),
                        "weight": 61,
                    }
                ],
                "deploy": [
                    {
                        "block": docker_run_cmd(
                            ["mkdir /home/$USER && useradd -u $UID -g $GID -d $HOME -m -s /bin/bash $USER"]
                        ),
                        "weight": 61,
                    }
                ],
            }
        }
    )



"""Test utility functions."""
from pydantic import ValidationError


def error_locations(ve: ValidationError) -> list[tuple[str, ...]]:
    """Return all the error locations from a ValidationError."""
    return [e["loc"] for e in ve.errors()]



"""Test the sources settings validation."""
from pydantic import ValidationError
from pytest import raises

from ou_container_builder.settings import WebApp

from .util import error_locations


def test_valid_web_app_settings():
    """Test that a valid sources configuration passes."""
    WebApp(
        path="test",
        options={
            "command": ["python", "-m", "http.server", "{port}"],
            "port": 80,
            "timeout": 120,
            "absolute_url": True,
            "launcher_entry": {},
        },
    )
    WebApp(
        path="test",
        options={
            "command": "python -m http.server {port}",
            "port": 80,
            "timeout": 120,
            "absolute_url": True,
            "launcher_entry": {},
        },
    )


def test_default_web_app_settings():
    """Test that the default web_app settings are correct."""
    settings = WebApp(path="test", options={"command": ["python", "-m", "http.server", "{port}"]})
    assert settings.path == "test"
    assert settings.options["command"] == ["python", "-m", "http.server", "{port}"]



"""Functionality for setting up the base server."""
import json
import os

from importlib.resources import open_text, files
from rich.progress import Progress

from ou_container_builder.state import State
from ou_container_builder.util import docker_copy_cmd, docker_run_cmd


CONTAINER_SCRIPT = """#!/bin/bash

mkdir -p $HOME
sudo chmod 775 $HOME

/var/lib/ou/python/system/bin/python /usr/share/ou/startup.py

if [[ ! -z "${JUPYTERHUB_API_TOKEN}" ]]; then
    export JUPYTERHUB_SINGLEUSER_APP='jupyter_server.serverapp.ServerApp'
    exec /var/lib/ou/python/system/bin/jupyterhub-singleuser
else
    exec /var/lib/ou/python/system/bin/jupyter server
fi
"""


def init(state: State) -> None:
    """Specify the base server configuration."""
    state.update(
        {
            "packages": {"pip": {"system": ["jupyter_server>=2.12,<3", "jupyterhub>=4.0.0,<5"]}},
            "jupyter_server_config": {
                "ServerApp": {"ip": "0.0.0.0", "port": 8888, "trust_xheaders": True},
                "ZMQChannelsWebsocketConnection": {"iopub_data_rate_limit": 10000000},
            },
        }
    )
    state.add_listener("server", server_listener)


def server_listener(state: State) -> None:
    """Update based on the server settings."""
    state.update({"jupyter_server_config": {"ServerApp": {"default_url": state["server"]["default_path"]}}})
    if state["server"]["access_token"] is not None:
        state.update({"jupyter_server_config": {"ServerApp": {"token": state["server"]["access_token"]}}})
    if state["server"]["wrapper_host"] is not None:
        state.update(
            {
                "jupyter_server_config": {
                    "ServerApp": {
                        "tornado_settings": {
                            "headers": {
                                "Content-Security-Policy": f"frame-ancestors 'self' {state['server']['wrapper_host']}"
                            }
                        }
                    }
                }
            }
        )


def generate(state: State, progress: Progress) -> None:
    """Generate the blocks for the deploy stage."""
    os.makedirs(os.path.join("ou-builder-build", "server"), exist_ok=True)
    with open(os.path.join("ou-builder-build", "server", "vce-start.sh"), "w") as out_f:
        out_f.write(CONTAINER_SCRIPT)
    with open(os.path.join("ou-builder-build", "server", "jupyter_server_config.json"), "w") as out_f:
        json.dump(state["jupyter_server_config"], out_f)
    with open(os.path.join("ou-builder-build", "server", "startup.py"), "w") as out_f:
        out_f.write(files("ou_container_builder.core.server").joinpath("startup.py").read_text())
    with open(os.path.join("ou-builder-build", "server", "99-chmod-homedir"), "w") as out_f:
        out_f.write(
            f"{state['image']['user']} ALL=(root) NOPASSWD: /usr/bin/chmod 775 /home/{state['image']['user']}/{state['module']['code']}-{state['module']['presentation']}\n"
        )
    state.update(
        {
            "output_blocks": {
                "deploy": [
                    {
                        "block": docker_copy_cmd(
                            os.path.join("ou-builder-build", "server", "vce-start.sh"), "/usr/bin/vce-start.sh"
                        ),
                        "weight": 181,
                    },
                    {
                        "block": docker_copy_cmd(
                            os.path.join("ou-builder-build", "server", "jupyter_server_config.json"),
                            "/usr/local/etc/jupyter/jupyter_server_config.json",
                        ),
                        "weight": 181,
                    },
                    {"block": docker_run_cmd(["chmod a+x /usr/bin/vce-start.sh"]), "weight": 182},
                    {
                        "block": docker_copy_cmd(
                            os.path.join("ou-builder-build", "server", "startup.py"), "/usr/share/ou/startup.py"
                        ),
                        "weight": 183,
                    },
                    {
                        "block": docker_copy_cmd(
                            os.path.join("ou-builder-build", "server", "99-chmod-homedir"), "/etc/sudoers.d"
                        ),
                        "weight": 184,
                    },
                ]
            }
        }
    )



"""Functionality for setting up the system environment variables."""
from rich.progress import Progress

from ou_container_builder.state import State


def init(state: State) -> None:
    """Unused."""
    state.update(
        {
            "environment": [
                {"name": "USER", "value": None},
                {"name": "UID", "value": "1000"},
                {"name": "GID", "value": "100"},
                {"name": "MODULE_CODE", "value": None},
                {"name": "MODULE_PRESENTATION", "value": None},
                {"name": "HOME", "value": "/home/$USER/$MODULE_CODE-$MODULE_PRESENTATION"},
                {"name": "SHELL", "value": "/bin/bash"},
            ]
        }
    )
    state.add_listener("image.user", update_environment)
    state.add_listener("module.code", update_environment)
    state.add_listener("module.presentation", update_environment)


def update_environment(state: State):
    """Update the environment setting based on the image and module states."""
    for env in state["environment"]:
        if env["name"] == "USER" and "image" in state and "user" in state["image"]:
            env["value"] = state["image"]["user"]
        elif env["name"] == "MODULE_CODE" and "module" in state and "code" in state["module"]:
            env["value"] = state["module"]["code"]
        elif env["name"] == "MODULE_PRESENTATION" and "module" in state and "presentation" in state["module"]:
            env["value"] = state["module"]["presentation"]


def generate(state: State, progress: Progress) -> None:
    """Generate the ENV blocks for the two stages."""
    state.update(
        {
            "output_blocks": {
                "build": [
                    {
                        "block": "\n".join(f'ENV {env["name"]}="{env["value"]}"' for env in state["environment"]),
                        "weight": 41,
                    }
                ],
                "deploy": [
                    {
                        "block": "\n".join(f'ENV {env["name"]}="{env["value"]}"' for env in state["environment"]),
                        "weight": 41,
                    }
                ],
            }
        }
    )



"""Functionality to handle installing apt packages."""
from rich.progress import Progress

from ou_container_builder.state import State
from ou_container_builder.util import docker_run_cmd


def init(state: State) -> None:
    """Unused."""
    pass


def generate(state: State, progress: Progress) -> None:
    """Generate the blocks to install all apt packages in the build and deploy stages."""
    state.update(
        {
            "output_blocks": {
                "build": [
                    {
                        "block": docker_run_cmd(
                            [
                                "DEBIAN_FRONTEND=noninteractive apt-get update -y",
                                "DEBIAN_FRONTEND=noninteractive apt-get dist-upgrade -y",
                                f"DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends {' '.join(state['packages']['apt']['core'])}",
                            ]
                        ),
                        "weight": 91,
                    },
                ],
                "deploy": [
                    {
                        "block": docker_run_cmd(
                            [
                                "DEBIAN_FRONTEND=noninteractive apt-get update -y",
                                "DEBIAN_FRONTEND=noninteractive apt-get dist-upgrade -y",
                                f"DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends {' '.join(state['packages']['apt']['core'])}",
                            ]
                        ),
                        "weight": 91,
                    },
                ],
            }
        }
    )
    if len(state["packages"]["apt"]["build"]) > 0:
        state.update(
            {
                "output_blocks": {
                    "build": [
                        {
                            "block": docker_run_cmd(
                                [
                                    "DEBIAN_FRONTEND=noninteractive apt-get update -y",
                                    "DEBIAN_FRONTEND=noninteractive apt-get dist-upgrade -y",
                                    f"DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends {' '.join(state['packages']['apt']['build'])}",
                                ]
                            ),
                            "weight": 141,
                        },
                    ],
                }
            }
        )
    if len(state["packages"]["apt"]["deploy"]) > 0:
        state.update(
            {
                "output_blocks": {
                    "deploy": [
                        {
                            "block": docker_run_cmd(
                                [
                                    "DEBIAN_FRONTEND=noninteractive apt-get update -y",
                                    "DEBIAN_FRONTEND=noninteractive apt-get dist-upgrade -y",
                                    f"DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends {' '.join(state['packages']['apt']['deploy'])}",
                                ]
                            ),
                            "weight": 141,
                        },
                    ],
                }
            }
        )



"""Functionality to install an iPython kernel."""
import os

from pydantic import BaseModel

from rich.progress import Progress

from ou_container_builder.state import State
from ou_container_builder.util import docker_copy_cmd

name = "ipykernel"
STARTUP_KERNEL_SCRIPT = """#!/bin/bash

export PATH=$PATH:$HOME/.local/bin
jupyter kernelspec remove -y python3
ipython kernel install --name "Python-3" --user
"""


class Options(BaseModel):
    """Options for the ipykernel pack."""

    pass


def init(state: State) -> None:
    """Initialise packs.ipykernel."""
    state.update({"environment": [{"name": "PYDEVD_DISABLE_FILE_VALIDATION", "value": "1"}]})
    state.update({"packages": {"pip": {"user": ["ipykernel"]}}})


def generate(state: State, progress: Progress) -> None:
    """Generate the kernel installation script."""
    os.makedirs(os.path.join("ou-builder-build", "ipykernel"), exist_ok=True)
    with open(os.path.join("ou-builder-build", "ipykernel", "110-activating-the-python-kernel"), "w") as out_f:
        out_f.write(STARTUP_KERNEL_SCRIPT)
    state.update(
        {
            "output_blocks": {
                "deploy": [
                    {
                        "block": docker_copy_cmd(
                            os.path.join("ou-builder-build", "ipykernel", "110-activating-the-python-kernel"),
                            "/usr/share/ou/startup.d/110-activating-the-python-kernel",
                        ),
                        "weight": 191,
                    }
                ]
            }
        }
    )



"""Script to prepare the version information for a release."""

import re

VERSION = "3.0.0b6"


def readlines(filename: str) -> list[str]:
    """Read a file as a list of lines."""
    with open(filename) as in_f:
        return in_f.readlines()


def writelines(filename: str, lines: list[str]) -> None:
    """Write a list of lines to a file."""
    with open(filename, "w") as out_f:
        out_f.write("".join(lines))


def update_version(filename: str, pattern: str, version: str) -> None:
    """Update the version in a file."""

    def replace_version(line: str) -> str:
        if re.match(pattern, line):
            return re.sub(pattern, fr"\g<1>{version}\g<2>", line)
        else:
            return line

    writelines(filename, map(replace_version, readlines(filename)))


update_version("docs/conf.py", r'(^release = ")[0-9]+\.[0-9]+\.[0-9]+(?:[ab][0-9]+)?("$)', VERSION)
update_version(
    "ou_container_builder/__about__.py", r'(^__version__ = ")[0-9]+\.[0-9]+\.[0-9]+(?:[ab][0-9]+)?("$)', VERSION
)
update_version("CHANGELOG.md", r"(## )Dev($)", VERSION)



"""Utility functions for working with Docker."""


def docker_run_cmd(commands: list[str]) -> str:
    """Generate a docker run command."""
    command_string = " && \\\n    ".join(commands)
    return f'RUN {command_string}'


def docker_copy_cmd(src: str, target: str, from_stage: str | None = None, chmod: str | None = None) -> str:
    """Generate a docker copy command."""
    parts = ["COPY"]
    if from_stage is not None:
        parts.append(f"--from={from_stage}")
    if chmod is not None:
        parts.append(f"--chmod={chmod}")
    parts.append(f'["{src}", "{target}"]')
    return " ".join(parts)



"""The main OU Container Builder commandline application.

Run ``ou-container-builder --help`` for help with the command-line parameters.
"""
from ou_container_builder.cli import app


if __name__ == "__main__":
    app()



"""Startup script run when the VCE launches."""

import asyncio
import os
import re
import tornado
import tornado.websocket

from asyncio import subprocess

HTML_TEMPLATE = """<!DOCTYPE html>
<html lang="en">
<head>
  <title>Your VCE is starting</title>
  <style>
    #status { position: absolute; left: 50%; top: 50%; max-width: 30rem; transform: translate(-50%,-50%); border-top-right-radius: 1rem; box-shadow: 0 0 #0000,var(--tw-ring-shadow, 0 0 #0000),0 10px 15px -3px rgba(0, 0, 0, 0.1), 0 4px 6px -4px rgba(0, 0, 0, 0.1); }
    h1 { font-size: 1rem; font-weight: bold; padding: 0.5rem 1rem; margin: 0; background-color: rgb(205 207 234); border-top-right-radius: 1rem; color: rgb(39 38 87); }
    #messages { max-height: 12rem; margin: 0; padding: 0.5rem 1rem; overflow: auto; border-left: 1px solid; border-bottom: 1px solid; border-right: 1px solid; border-color: rgb(214 214 214); }
    #messages > li { list-style-type: none; display: flex; flex-direction: row; align-items: center; margin-bottom: 0.5rem; }
    .msg { flex: 1 1 auto; padding-right: 1rem; }
    .icon { flex: 0 0 auto; display: block; width: 1.5rem; height: 1.5rem; color: rgb(39 38 87); }
    .icon svg { display: block; width: 100%; height: 100%; fill: currentColor; }
    .icon.busy svg { animation: spin 1s linear infinite; }
    .icon.failure { color: rgb(168 12 12); }

    @keyframes spin {
      from {
        transform: rotate(0deg);
      }
      to {
        transform: rotate(360deg);
      }
    }
  </style>
</head>
<body>
  <div id="status" role="status">
    <h1 id="header" tabindex="-1">Your VCE is starting</h1>
    <ul id="messages" aria-live="true">
    </ul>
  </div>
  <script>
    const header = document.querySelector("#header");
    header.focus();

    const messages = document.querySelector("#messages");
    const ws = new WebSocket(((window.location.protocol === "https:") ? "wss://" : "ws://") + window.location.host + "{JUPYTERHUB_SERVICE_PREFIX}ocl-status");
    addNewProgress({text: "Preparing startup"});

    /**
     * Handle websocket messages.
     */
    ws.addEventListener("message", (msg) => {
      const data = JSON.parse(msg.data);
      if (data.type === "start") {
        completeProgress({status: "success"});
        addNewProgress(data);
      } else {
        completeProgress(data);
      }
    });

    /**
     * Handle closing the websocket connection.
     */
    ws.addEventListener("close", () => {
      completeProgress({status: "success"});
      addNewProgress({text: "Waiting for the interface to be ready..."});
      setTimeout(checkReady, 5000);
    });

    /**
     * Add a new active progress element
     */
    function addNewProgress(progress) {
      const item = document.createElement("li");
      item.innerHTML = '<span class="msg">' + progress.text + '</span><span class="icon busy"><svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 24 24"><title>working</title><path d="M12,6V9L16,5L12,1V4A8,8 0 0,0 4,12C4,13.57 4.46,15.03 5.24,16.26L6.7,14.8C6.25,13.97 6,13 6,12A6,6 0 0,1 12,6M18.76,7.74L17.3,9.2C17.74,10.04 18,11 18,12A6,6 0 0,1 12,18V15L8,19L12,23V20A8,8 0 0,0 20,12C20,10.43 19.54,8.97 18.76,7.74Z" /></svg></span>';
      messages.appendChild(item);
      setTimeout(() => {
        item.scrollIntoView();
      }, 0);
    }

    /**
     * Update the active progress to either success or failure.
     */
    function completeProgress(progress) {
      const current = document.querySelector(".busy");
      if (current !== null) {
        current.classList.remove("busy");
        if (progress.status === "success") {
          current.innerHTML = '<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 24 24"><title>Completed successfully</title><path d="M9,20.42L2.79,14.21L5.62,11.38L9,14.77L18.88,4.88L21.71,7.71L9,20.42Z" /></svg>';
          current.classList.add("success");
        } else {
          current.innerHTML = '<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 24 24"><title>close-thick</title><path d="M20 6.91L17.09 4L12 9.09L6.91 4L4 6.91L9.09 12L4 17.09L6.91 20L12 14.91L17.09 20L20 17.09L14.91 12L20 6.91Z" /></svg>';
          current.classList.add("failure");
        }
      }
    }

    /**
     * Check whether the UI is ready and reload if it is.
     */
    async function checkReady() {
      try {
        const response = await window.fetch(window.location.href);
        if (Math.floor(response.status / 100) === 4 || Math.floor(response.status / 100) === 5) {
          setTimeout(checkReady, 5000);
        } else {
          window.location.reload();
        }
      } catch(e) {
        setTimeout(() => {
          window.location.reload();
        }, 5000);
      }
    }
  </script>
</body>
</html>
"""


class IndexRequestHandler(tornado.web.RequestHandler):
    """Handler for any URL to return the HTML."""

    def get(self: "IndexRequestHandler") -> None:
        """Send the HTML template to the user."""
        self.set_header("Content-Type", "text/html")
        self.set_header("Cache-Control", "no-cache, no-store")
        prefix = os.environ["JUPYTERHUB_SERVICE_PREFIX"] if "JUPYTERHUB_SERVICE_PREFIX" in os.environ else "/"
        self.write(HTML_TEMPLATE.replace("{JUPYTERHUB_SERVICE_PREFIX}", prefix))


class ProgressHandler(tornado.websocket.WebSocketHandler):
    """Handler for the progress Websocket."""

    def initialize(self: "ProgressHandler", queue: asyncio.queues.Queue) -> None:
        """Initialise the handler with the queue to use."""
        self._queue = queue

    def open(self: "ProgressHandler"):
        """Open the Websocket connection and start the queue tracking."""
        asyncio.create_task(self.track_progress())

    async def track_progress(self: "ProgressHandler"):
        """Track the progress queue and send it on to the websocket."""
        while True:
            msg = await self._queue.get()
            self.write_message(msg)


async def main():
    """Run the progress server and the startup scripts."""
    done_event = asyncio.Event()
    queue = asyncio.queues.Queue()

    async def run_startup_scripts():
        """Task to run the individual startup scripts."""
        print("Preparing startup")
        await asyncio.sleep(1)
        filenames = os.listdir("/usr/share/ou/startup.d")
        filenames.sort()
        print("Preparing startup complete")
        for filename in filenames:
            tokens = filename.split('-')
            if len(tokens) > 0:
                if re.match("^[0-9]+$", tokens[0]):
                    tokens = tokens[1:]
            if len(tokens) > 0:
                tokens[0] = tokens[0].title()
                message = " ".join(tokens)
            else:
                message = "Unknown activity"
            print(message, flush=True)
            await queue.put({"type": "start", "text": message})
            process = await subprocess.create_subprocess_exec("bash", f"/usr/share/ou/startup.d/{filename}")
            await process.wait()
            if process.returncode != 0:
                await queue.put({"type": "done", "status": "failure"})
            else:
                await queue.put({"type": "done", "status": "success"})
                print(f"{message} complete", flush=True)
            await asyncio.sleep(0.5)
        done_event.set()

    application = tornado.web.Application(
        [("/ocl-status", ProgressHandler, {"queue": queue}), ("/.*", IndexRequestHandler)]
    )
    application.listen(8888)
    asyncio.get_event_loop().create_task(run_startup_scripts())
    await done_event.wait()


if __name__ == "__main__":
    asyncio.run(main())



"""Clean-up function."""
import os
import shutil

from ou_container_builder.cli.base import app


@app.command()
def clean():
    """Clean the build files."""
    if os.path.exists("ou-builder-build"):
        shutil.rmtree("ou-builder-build")
    if os.path.exists("Dockerfile"):
        os.unlink("Dockerfile")



"""Functionality for setting up additional apt sources in both stages."""
from rich.progress import Progress

from ou_container_builder.state import State
from ou_container_builder.util import docker_run_cmd


def init(state: State) -> None:
    """Unused."""
    state.add_listener("sources.apt", apt_sources_listener)


def apt_sources_listener(state: State) -> None:
    """If additional sources are configured, add required core packages."""
    if len(state["sources"]["apt"]) > 0:
        state.update({"packages": {"apt": {"core": ["gnupg"]}}})


def generate(state: State, progress: Progress) -> None:
    """Generate the commands to download signing keys and create source list files."""
    if len(state["sources"]["apt"]) > 0:
        commands = []
        for source in state["sources"]["apt"]:
            if source["dearmor"]:
                commands.append(
                    f'curl -fsSL "{source["key_url"]}" | gpg --dearmor -o "/usr/share/keyrings/{source["name"]}.gpg"'
                )
            else:
                commands.append(f'curl -fsSL "{source["key_url"]}" > "/usr/share/keyrings/{source["name"]}.gpg"')
            commands.append(
                f'echo "deb [signed-by=/usr/share/keyrings/{source["name"]}.gpg] {source["deb"]["url"]} {source["deb"]["distribution"]} {source["deb"]["component"]}" > "/etc/apt/sources.list.d/{source["name"]}.list"'
            )
        state.update(
            {
                "output_blocks": {
                    "build": [
                        {
                            "block": docker_run_cmd(commands),
                            "weight": 121,
                        }
                    ],
                    "deploy": [
                        {
                            "block": docker_run_cmd(commands),
                            "weight": 121,
                        }
                    ],
                }
            }
        )



"""Functionality for setting up rust in the build stage."""
from rich.progress import Progress

from ou_container_builder.state import State
from ou_container_builder.util import docker_run_cmd


def init(state: State) -> None:
    """Specify the base packages to install."""
    state.update({"packages": {"apt": {"core": ["curl"]}}})


def generate(state: State, progress: Progress) -> None:
    """Generate the FROM blocks for the two stages."""
    state.update(
        {
            "output_blocks": {
                "build": [
                    {
                        "block": docker_run_cmd(
                            ["curl --proto '=https' --tlsv1.2 -sSf https://sh.rustup.rs | sh -s -- -y"]
                        ),
                        "weight": 101,
                    },
                ],
            }
        }
    )



"""Functionality for setting up the build arguments."""
from rich.progress import Progress

from ou_container_builder.state import State


def init(state: State) -> None:
    """Set the default ARGs."""
    state.update({"args": []})


def generate(state: State, progress: Progress) -> None:
    """Generate the ARG blocks for the two stages."""
    state.update(
        {
            "output_blocks": {
                "build": [
                    {
                        "block": "\n".join(
                            f'ARG {arg["name"]}="{arg["value"]}"' if arg["value"] is not None else f"ARG {arg['name']}"
                            for arg in state["args"]
                        ),
                        "weight": 21,
                    }
                ],
                "deploy": [
                    {
                        "block": "\n".join(
                            f'ARG {arg["name"]}="{arg["value"]}"' if arg["value"] is not None else f"ARG {arg['name']}"
                            for arg in state["args"]
                        ),
                        "weight": 21,
                    }
                ],
            }
        }
    )



"""Version CLI command."""
from ou_container_builder.__about__ import __version__
from ou_container_builder.cli.base import app


@app.command()
def version():
    """Print the installed version."""
    print(__version__)  # noqa: T201



"""Test the sources settings validation."""
from pydantic import ValidationError
from pytest import raises

from ou_container_builder.settings import AptDebLine, AptSource, Sources

from .util import error_locations


def test_valid_sources_settings():
    """Test that a valid sources configuration passes."""
    Sources(apt=[])


def test_default_sources_settings():
    """Test that the default sources settings are correct."""
    settings = Sources()
    assert settings.apt == []


def test_valid_apt_source_settings():
    """Test that a valid apt source configuration passes."""
    settings = AptSource(
        name="test",
        key_url="https://example.com/key-url",
        dearmor=False,
        deb={"url": "https://example.com/repo", "distribution": "test", "component": "main"},
    )
    assert settings.dearmor is False


def test_default_apt_source_settings():
    """Test that the default apt source settings are correct."""
    settings = AptSource(
        name="test",
        key_url="https://example.com/key-url",
        deb={"url": "https://example.com/repo", "distribution": "test", "component": "main"},
    )
    assert settings.dearmor is True


def test_invalid_apt_source_key_url_fails():
    """Test that an invalid key_url fails."""
    with raises(ValidationError) as e_info:
        AptSource(
            name="test",
            key_url="nowhere",
            deb={"url": "https://example.com/repo", "distribution": "test", "component": "main"},
        )
    assert ("key_url",) in error_locations(e_info.value)


def test_valid_apt_deb_line_settings():
    """Test that a valid apt source configuration passes."""
    AptDebLine(url="https://example.com/repo", distribution="test", component="main")


def test_invalid_apt_deb_line_url_fails():
    """Test that an invalid key_url fails."""
    with raises(ValidationError) as e_info:
        AptDebLine(url="nowhere", distribution="test", component="main")
    assert ("url",) in error_locations(e_info.value)



"""Test the server settings validation."""
from ou_container_builder.settings import Server


def test_default_server_settings():
    """Test that the default server configuration passes."""
    settings = Server()
    assert settings.default_path == "/"
    assert settings.access_token is None
    assert settings.wrapper_host is None


def test_valid_server_settings():
    """Test that a valid server configuration passes."""
    settings = Server(default_path="/lab", access_token="test", wrapper_host="*.open.ac.uk")
    assert settings.default_path == "/lab"
    assert settings.access_token == "test"
    assert settings.wrapper_host == "*.open.ac.uk"



"""Functionality to install CodeServer."""
import os
import shutil

from importlib.resources import open_text
from pydantic import BaseModel
from typing import Literal

from rich.progress import Progress

from ou_container_builder.util import docker_copy_cmd, docker_run_cmd
from ou_container_builder.state import State


name = "code_server"


class Options(BaseModel):
    """Options for the CodeServer pack."""

    version: str = "4.19.1"
    extensions: list[str] = []


def init(state: State) -> None:
    """Initialise packs.code_server."""
    state.update(
        {
            "sources": {
                "apt": [
                    {
                        "name": "nodesource",
                        "key_url": "https://deb.nodesource.com/gpgkey/nodesource-repo.gpg.key",
                        "dearmor": True,
                        "deb": {
                            "url": "https://deb.nodesource.com/node_20.x",
                            "distribution": "nodistro",
                            "component": "main",
                        },
                    }
                ]
            },
            "packages": {"apt": {"deploy": ["nodejs", "yarn"]}},
            "web_apps": [
                {
                    "path": "code-server",
                    "options": {
                        "command": [
                            "code-server",
                            "--auth",
                            "none",
                            "--disable-update-check",
                            "--bind-addr",
                            "0.0.0.0",
                            "--port",
                            "{port}",
                        ],
                        "timeout": 60,
                        "new_browser_tab": False,
                        "launcher_entry": {
                            "title": "VS Code",
                            "icon_path": "/usr/share/ou/packs/code_server/vscode.svg",
                        },
                    },
                }
            ],
        }
    )


def generate(state: State, progress: Progress) -> None:
    """Generate the docker blocks to install Code Server."""
    if os.path.exists(os.path.join("ou-builder-build", "code_server")):
        shutil.rmtree(os.path.join("ou-builder-build", "code_server"))
    os.makedirs(os.path.join("ou-builder-build", "code_server"), exist_ok=True)
    with open(os.path.join("ou-builder-build", "code_server", "vscode.svg"), "w") as out_f:
        with open_text("ou_container_builder.packs.code_server", "vscode.svg") as in_f:
            out_f.write(in_f.read())
    version = state["packs"]["code_server"]["version"]
    state.update(
        {
            "output_blocks": {
                "build": [
                    {
                        "block": docker_run_cmd(
                            [
                                f"curl -L --output /tmp/code-server.tar.gz https://github.com/coder/code-server/releases/download/v{version}/code-server-{version}-linux-amd64.tar.gz",
                                "tar -zxf /tmp/code-server.tar.gz --directory /opt",
                            ]
                        ),
                        "weight": 1041,
                    }
                ],
                "deploy": [
                    {
                        "block": docker_copy_cmd(
                            f"/opt/code-server-{version}-linux-amd64",
                            f"/opt/code-server-{version}-linux-amd64",
                            from_stage="base",
                        ),
                        "weight": 1041,
                    },
                    {
                        "block": docker_copy_cmd(
                            os.path.join("ou-builder-build", "code_server", "vscode.svg"),
                            "/usr/share/ou/packs/code_server/vscode.svg",
                        ),
                        "weight": 1042,
                    },
                    {
                        "block": docker_run_cmd(
                            [f"ln -s /opt/code-server-{version}-linux-amd64/bin/code-server /usr/local/bin"]
                        ),
                        "weight": 1043,
                    },
                ],
            }
        }
    )
    if len(state["packs"]["code_server"]["extensions"]) > 0:
        for extension in state["packs"]["code_server"]["extensions"]:
            state.update(
                {
                    "output_blocks": {
                        "deploy": [
                            {"block": f"USER {state['image']['user']}", "weight": 1044},
                            {"block": docker_run_cmd([f"code-server --install-extension {extension}"]), "weight": 1045},
                            {"block": "USER root", "weight": 1046},
                        ]
                    }
                }
            )



"""Functionality for running scripts in the build and live phase."""
# TODO: Validation tests needed
import os
import shutil

from rich.progress import Progress

from ou_container_builder.state import State
from ou_container_builder.util import docker_copy_cmd, docker_run_cmd


def init(state: State) -> None:
    """Unused."""
    pass


def generate(state: State, progress: Progress) -> None:
    """Generate the FROM blocks for the two stages."""
    if os.path.exists(os.path.join("ou-builder-build", "scripts")):
        shutil.rmtree(os.path.join("ou-builder-build", "scripts"))
    os.makedirs(os.path.join("ou-builder-build", "scripts", "startup"), exist_ok=True)
    task = progress.add_task("Generating custom scripts", total=len(state["scripts"]))
    for idx, script in enumerate(state["scripts"]):
        if script["stage"] == "build":
            state.update(
                {"output_blocks": {"build": [{"block": docker_run_cmd(script["commands"]), "weight": 201 + idx}]}}
            )
        elif script["stage"] == "deploy":
            state.update(
                {"output_blocks": {"deploy": [{"block": docker_run_cmd(script["commands"]), "weight": 201 + idx}]}}
            )
        elif script["stage"] == "startup":
            if not script["commands"][0].startswith("#!"):
                script["commands"].insert(0, "#!/bin/bash")
            filename = f"{201 + idx}-{script['name'].lower().replace(' ', '-')}"
            with open(os.path.join("ou-builder-build", "scripts", "startup", filename), "w") as out_f:
                for line in script["commands"]:
                    out_f.write(line)
                    out_f.write("\n")
            state.update(
                {
                    "output_blocks": {
                        "deploy": [
                            {
                                "block": docker_copy_cmd(
                                    os.path.join("ou-builder-build", "scripts", "startup", filename),
                                    f"/usr/share/ou/startup.d/{filename}",
                                ),
                                "weight": 201 + idx,
                            }
                        ]
                    }
                }
            )
        progress.update(task, advance=1)
    progress.update(task, visible=False)



"""Test the content settings validation."""
from pydantic import ValidationError
from pytest import raises

from ou_container_builder.settings import Content

from .util import error_locations


def test_valid_content_settings():
    """Test that a valid content configurations pass."""
    Content(source="test", target="/var/lib/test", overwrite="always")
    Content(source="test", target="/var/lib/test", overwrite="never")


def test_default_settings():
    """Test that the default settings are set correctly."""
    settings = Content(source="test", overwrite="always")
    assert settings.target == ""


def test_invalid_overwrite_mode():
    """Test that an invalid overwrite setting fails."""
    with raises(ValidationError) as e_info:
        Content(source="test", overwrite="sometimes")



"""Test the module settings validation."""
from pydantic import ValidationError
from pytest import raises

from ou_container_builder.settings import Module

from .util import error_locations


def test_valid_module_settings():
    """Test that a valid module configuration passes."""
    Module(code="TEST001", presentation="23J")


def test_missing_module_code():
    """Test that a module configuration without a code fails."""
    with raises(ValidationError) as e_info:
        Module(presentation="23J")
    assert ("code",) in error_locations(e_info.value)


def test_missing_module_presentation():
    """Test that a module configuration without a presentation fails."""
    with raises(ValidationError) as e_info:
        Module(code="TEST001")
    assert ("presentation",) in error_locations(e_info.value)



"""Functionality to install JupyterLab."""
from pydantic import BaseModel
from typing import Literal

from rich.progress import Progress

from ou_container_builder.state import State

name = "jupyterlab"


class Options(BaseModel):
    """Options for the JupyterLab pack."""

    version: Literal[3] | Literal[4] = 4


def init(state: State) -> None:
    """Initialise packs.jupyterlab."""
    if state["packs"]["jupyterlab"]["version"] == 3:
        state.update({"packages": {"pip": {"system": ["jupyterlab>=3.4.3,<4"]}}})
    elif state["packs"]["jupyterlab"]["version"] == 4:
        state.update({"packages": {"pip": {"system": ["jupyterlab>=4.0.2,<5"]}}})


def generate(state: State, progress: Progress) -> None:
    """Unused."""
    pass



"""Test the version command."""
from typer import Typer
from typer.testing import CliRunner

from ou_container_builder.__about__ import __version__


def test_current_version(runner: CliRunner, app: Typer):
    """Test that the current version is printed."""
    result = runner.invoke(app, ["version"])
    assert result.exit_code == 0
    assert result.stdout == f"{__version__}\n"



"""Functionality to install Notebook."""

from pydantic import BaseModel
from typing import Literal

from rich.progress import Progress

from ou_container_builder.state import State

name = "notebook"


class Options(BaseModel):
    """Options for the Notebook pack."""

    version: Literal[7] = 7


def init(state: State) -> None:
    """Initialise packs.notebook."""
    if state["packs"]["notebook"]["version"] == 7:
        state.update({"packages": {"pip": {"system": ["notebook>=7.1.1,<8"]}}})


def generate(state: State, progress: Progress) -> None:
    """Unused."""
    pass



"""Functionality for setting up additional apt sources in both stages."""
import os
import shutil

from rich.progress import Progress

from ou_container_builder.state import State
from ou_container_builder.util import docker_copy_cmd


def init(state: State) -> None:
    """Add a services listener."""
    state.add_listener("services", services_listener)


def services_listener(state: State) -> None:
    """If services are configured, add required core packages."""
    if len(state["services"]) > 0:
        state.update({"packages": {"apt": {"deploy": ["sudo"]}}})


def generate(state: State, progress: Progress) -> None:
    """Generate the commands to start services."""
    if os.path.exists(os.path.join("ou-builder-build", "services")):
        shutil.rmtree(os.path.join("ou-builder-build", "services"))
    os.makedirs(os.path.join("ou-builder-build", "services", "sudoers"), exist_ok=True)
    os.makedirs(os.path.join("ou-builder-build", "services", "startup"), exist_ok=True)
    if len(state["services"]) > 0:
        for idx, service in enumerate(state["services"]):
            with open(os.path.join("ou-builder-build", "services", "sudoers", f"99-{service}"), "w") as out_f:
                out_f.write(
                    f"""{state['image']['user']} ALL=(root) NOPASSWD: /usr/sbin/service {service} start
{state['image']['user']} ALL=(root) NOPASSWD: /usr/sbin/service {service} restart
{state['image']['user']} ALL=(root) NOPASSWD: /usr/sbin/service {service} stop
"""
                )
            with open(
                os.path.join("ou-builder-build", "services", "startup", f"{50 + idx:03d}-starting-{service}"), "w"
            ) as out_f:
                out_f.write(
                    f"""#!/bin/bash

sudo /usr/sbin/service {service} start
"""
                )

        state.update(
            {
                "output_blocks": {
                    "deploy": [
                        {
                            "block": docker_copy_cmd(
                                os.path.join("ou-builder-build", "services", "sudoers"), "/etc/sudoers.d", chmod="0440"
                            ),
                            "weight": 241,
                        },
                        {
                            "block": docker_copy_cmd(
                                os.path.join("ou-builder-build", "services", "startup"), "/usr/share/ou/startup.d"
                            ),
                            "weight": 241,
                        },
                    ]
                }
            }
        )



"""Functionality for setting up the user in both stages."""
from rich.progress import Progress

from ou_container_builder.state import State
from ou_container_builder.util import docker_run_cmd


def init(state: State) -> None:
    """Add a listener for the web_apps key."""
    state.add_listener("web_apps", activate_webapps)


def activate_webapps(state: State) -> None:
    """Activate the jupyter-server-proxy and set the config."""
    if len(state["web_apps"]) > 0:
        state.update({"packages": {"pip": {"system": ["jupyter-server-proxy>=4.1.0,<5"]}}})
        state.update(
            {
                "jupyter_server_config": {
                    "ServerProxy": {
                        "servers": dict([(web_app["path"], web_app["options"]) for web_app in state["web_apps"]])
                    }
                }
            }
        )


def generate(state: State, progress: Progress) -> None:
    """Unused."""
    pass



"""Build function."""
import subprocess
import typer

from typing import Annotated, Optional


from ou_container_builder.cli.base import app
from ou_container_builder.cli.generate import generate
from ou_container_builder.cli.clean import clean


@app.command()
def build(
    tag: Annotated[Optional[list[str]], typer.Option(help="Docker tags to use")] = None,
    cache: Annotated[Optional[bool], typer.Option(help="Cache intermediate layers (true)")] = True,
) -> None:
    """Build the container image."""
    generate()
    cmd = ["buildah", "build", "--jobs", "2"]
    if tag is not None:
        for t in tag:
            cmd.extend(["--tag", t])
    if cache:
        cmd.extend(["--layers"])
    cmd.append(".")
    process = subprocess.run(cmd)
    if process.returncode == 0:
        clean()



"""Test running without a command."""
from typer import Typer
from typer.testing import CliRunner


def test_no_command(runner: CliRunner, app: Typer):
    """Test that not providing a command fails."""
    result = runner.invoke(app)
    assert result.exit_code == 2
    assert "Usage:" in result.stdout


def test_empty_command(runner: CliRunner, app: Typer):
    """Test that providing an empty command list fails."""
    result = runner.invoke(app, [])
    assert result.exit_code == 2
    assert "Usage:" in result.stdout


def test_invalid_command(runner: CliRunner, app: Typer):
    """Test that providing an invalid command fails."""
    result = runner.invoke(app, ["not-exist"])
    assert result.exit_code == 2
    assert "Usage:" in result.stdout



"""Functionality for setting up the base system."""
from rich.progress import Progress

from ou_container_builder.state import State


def init(state: State) -> None:
    """Specify the base packages to install."""
    state.update(
        {
            "packages": {
                "apt": {
                    "core": ["gcc", "build-essential"],
                    "build": [],
                    "deploy": [
                        "libcurl3-gnutls",
                        "libcurl3-gnutls-dev",
                        "gnutls-dev",
                        "sudo",
                        "tini",
                    ],
                },
                "pip": {"system": ["pycurl"], "user": ["pycurl"]},
            }
        }
    )


def generate(state: State, progress: Progress) -> None:
    """Generate the FROM blocks for the two stages."""
    state.update(
        {
            "output_blocks": {
                "build": [
                    {"block": "# ###########\n# Build Stage\n# ###########\n\n", "weight": 0},
                    {"block": f"FROM {state['image']['base']} as base", "weight": 1},
                    {"block": "USER root", "weight": 2},
                ],
                "deploy": [
                    {"block": "\n# ############\n# Deploy Stage\n# ############\n\n", "weight": 0},
                    {"block": f"FROM {state['image']['base']} as deploy", "weight": 1},
                    {"block": "USER root", "weight": 2},
                    {"block": "USER $USER\nWORKDIR $HOME\nEXPOSE 8888", "weight": 9001},
                    {"block": 'ENTRYPOINT ["tini", "-g", "--"]\nCMD ["/usr/bin/vce-start.sh"]', "weight": 9002},
                ],
            }
        }
    )



"""Configuration settings."""
import shlex
import sys

from copy import deepcopy
from pydantic import BaseModel, Field, HttpUrl, ValidationError, constr, model_validator, create_model
from rich import print as cli_print
from typing import Literal, Any, Optional
from typing_extensions import Annotated
from yaml import safe_load

from ou_container_builder import packs


class Module(BaseModel):
    """Settings for the module configuration."""

    code: constr(min_length=1)
    presentation: constr(min_length=1)


class Image(BaseModel):
    """Settings for the Docker image configuration."""

    base: constr(min_length=1) = "python:3.11-bookworm"
    user: constr(min_length=1) = "ou"


class EnvironmentVariable(BaseModel):
    """Settings for additional environment variables."""

    name: constr(min_length=1)
    value: str = ""


class Arg(BaseModel):
    """Docker argument."""

    name: constr(min_length=1)
    value: str | None = None


class Server(BaseModel):
    """Settings for the core server configuration."""

    default_path: constr(min_length=1) = "/"
    access_token: str | None = None
    wrapper_host: str | None = None


class Content(BaseModel):
    """Settings for the content configuration."""

    source: constr(min_length=1)
    target: str = ""
    overwrite: Literal["always"] | Literal["never"]


class AptDebLine(BaseModel):
    """Settings for an APT deb line."""

    url: HttpUrl
    distribution: constr(min_length=1)
    component: constr(min_length=1)


class AptSource(BaseModel):
    """Settings for a single APT source."""

    name: constr(min_length=1)
    key_url: HttpUrl
    dearmor: bool = True
    deb: AptDebLine


class Sources(BaseModel):
    """Settings for the additional sources configuration."""

    apt: list[AptSource] = []


class AptPackageLists(BaseModel):
    """Settings for a list of apt packages for build and deploy stages."""

    core: list[constr(min_length=1)] = []
    build: list[constr(min_length=1)] = []
    deploy: list[constr(min_length=1)] = []

    @model_validator(mode="before")
    @classmethod
    def add_stages(cls: "AptPackageLists", data: Any) -> Any:
        """If only a list of packages is given, convert that to the stage structure."""
        if isinstance(data, list):
            return {"build": deepcopy(data), "deploy": deepcopy(data)}
        return data


class PipPackageEntry(BaseModel):
    """Settings for a single entry for a pip installation.

    Can either represent a single file or a requirements file.
    """

    name: constr(min_length=1)
    type: Optional[Literal["package"] | Literal["requirements.txt"]] = "package"

    @model_validator(mode="before")
    @classmethod
    def convert_simple_packages(cls: "PipPackageLists", data: Any) -> Any:
        """If only a list of packages is given, convert that to the target structure."""
        if isinstance(data, str):
            return {"name": data, "type": "package"}
        return data


class PipPackageLists(BaseModel):
    """Settings for a list of pip packages for system and user targets."""

    system: list[PipPackageEntry] = []
    user: list[PipPackageEntry] = []

    @model_validator(mode="before")
    @classmethod
    def add_targets(cls: "PipPackageLists", data: Any) -> Any:
        """If only a list of packages is given, convert that to the target structure."""
        if isinstance(data, list):
            return {
                "system": [{"name": pkg, "type": "package"} for pkg in data],
                "user": [{"name": pkg, "type": "package"} for pkg in data],
            }
        return data


class Packages(BaseModel):
    """Settings for the packages configuration."""

    apt: AptPackageLists = AptPackageLists()
    pip: PipPackageLists = PipPackageLists()


class WebApp(BaseModel):
    """Settings for a single web application configuration."""

    path: constr(min_length=1)
    options: dict

    @model_validator(mode="before")
    @classmethod
    def convert_command_string_to_list(cls: "WebApp", data: Any) -> Any:
        """Convert command strings to lists using shlex."""
        if (
            isinstance(data, dict)
            and "options" in data
            and "command" in data["options"]
            and isinstance(data["options"]["command"], str)
        ):
            data["options"]["command"] = shlex.split(data["options"]["command"])
        return data


class BuildScript(BaseModel):
    """A script to be run during the build in the build or deploy stage."""

    stage: Literal["build"] | Literal["deploy"]
    commands: list[str]


class StartupShutdownScript(BaseModel):
    """A script to be run at startup or shutdown time."""

    stage: Literal["startup"]
    name: str
    commands: list[str]


PacksModel = create_model(
    "PacksModel", **dict([(key, (value.Options | None, None)) for key, value in packs.EXTENSION_PACKS.items()])
)


class OutputBlock(BaseModel):
    """Settings for a single block in the generated Dockerfile."""

    block: str
    weight: int = 999

    @model_validator(mode="before")
    @classmethod
    def convert_string_block(cls: "OutputBlock", data: Any) -> Any:
        """Convert string blocks into the dictionary structure."""
        if isinstance(data, str):
            return {"block": data}
        return data


class OutputBlocks(BaseModel):
    """Blocks for the generated Dockerfile."""

    build: list[OutputBlock] = []
    deploy: list[OutputBlock] = []


class Settings(BaseModel):
    """Application Settings."""

    version: Annotated[Literal["3"], Field(validate_default=True)] = "2"
    module: Module
    image: Image = Image()
    environment: list[EnvironmentVariable] = []
    args: list[Arg] = []
    server: Server = Server()
    content: list[Content] = []
    sources: Sources = Sources()
    packages: Packages = Packages()
    web_apps: list[WebApp] = []
    services: list[constr(min_length=1)] = []
    scripts: list[BuildScript | StartupShutdownScript] = []
    packs: PacksModel = {}
    jupyter_server_config: dict = {}
    output_blocks: OutputBlocks = OutputBlocks()


def load_settings() -> dict:
    """Load the settings from the ContainerConfig.yaml."""
    try:
        with open("ContainerConfig.yaml") as in_f:
            return Settings(**safe_load(in_f)).model_dump()
    except ValidationError as ve:
        cli_print("Unfortunately your ContainerConfig.yaml is not valid:\n")
        for err in ve.errors():
            cli_print(f"* [red]{'.'.join(err['loc'])}[/red]: {err['msg']}")
        sys.exit(1)



"""Builder state handling."""
from collections.abc import Mapping, Iterator
from copy import deepcopy

from typing import Any, Callable


def merge_settings(base: dict, new: dict) -> dict:
    """Return a new dictionary created by merging the settings from ``new`` into ``base``.

    :param base: The base dictionary to merge into
    :type base: ``dict``
    :param new: The new dictionary to merge
    :type new: ``dict``
    :return: A new merged dictionary
    :rtype: ``dict``
    """
    result = {}
    for base_key, base_value in list(base.items()):
        if base_key not in new:
            result[base_key] = deepcopy(base_value)
        else:
            if isinstance(base_value, list):
                result[base_key] = list(base_value + new[base_key])
            elif isinstance(base_value, dict):
                result[base_key] = merge_settings(base_value, new[base_key])
            else:
                result[base_key] = new[base_key]
    for new_key, new_value in list(new.items()):
        if new_key not in base:
            result[new_key] = deepcopy(new_value)
    return result


class State(Mapping):
    """The State represents the state of the builder and provides support for observing changes."""

    def __init__(self) -> None:
        """Initialise the State."""
        self._state = {}
        self._listeners: dict[str, Callable[["State", Any]]] = {}

    def update(self: "State", settings: dict) -> None:
        """Update the `State` with the given `settings`."""
        self._state = merge_settings(self._state, settings)
        self._notify(settings)

    def add_listener(self: "State", prefix: str, callback: Callable[["State", Any], None]):
        """Add a listener that is notified when the `State` changes at the location `prefix`."""
        if prefix in self._listeners:
            self._listeners[prefix].append(callback)
        else:
            self._listeners[prefix] = [callback]

    def _notify(self: "State", settings: dict) -> None:
        """Notify all listeners of a change."""

        def walk(current: dict, path: list[str] | None = None):
            """Walk the updated settings and send out notifications."""
            if path is None:
                path = []
            for key, value in current.items():
                prefix = ".".join(path + [key])
                if prefix in self._listeners:
                    for callback in self._listeners[prefix]:
                        callback(self)
                if isinstance(value, dict):
                    walk(value, path + [key])

        walk(settings)

    def __getitem__(self: "State", key: str) -> Any:
        """Retrieve an item from the `State`."""
        return self._state[key]

    def __len__(self: "State") -> int:
        """Return the length of this `State`."""
        return len(self._state)

    def __iter__(self: "State") -> Iterator:
        """Iterate over the items of this `State`."""
        return self._state.__iter__()



"""Functionality for copying content into the image."""
import os
import shutil
import subprocess

from rich.progress import Progress

from ou_container_builder.state import State
from ou_container_builder.util import docker_copy_cmd


def init(state: State) -> None:
    """Unused."""
    pass


def generate(state: State, progress: Progress) -> None:
    """Copy the content files."""
    content_task = progress.add_task("Compressing content to distribute...", target=len(state["content"]))
    os.makedirs(os.path.join("ou-builder-build", "content", "archives"), exist_ok=True)
    unpack_commands = []
    for idx, content in enumerate(state["content"]):
        if content["target"].startswith("/"):
            state.update(
                {
                    "output_blocks": {
                        "deploy": [{"block": docker_copy_cmd(content["source"], content["target"]), "weight": 181}]
                    }
                }
            )
        else:
            if content["source"].endswith("/"):
                content["source"] = content["source"][:-1]
            if os.path.isfile(content["source"]):
                shutil.copyfile(
                    content["source"], os.path.join("ou-builder-build", "content", "archives", f"content-{idx}")
                )
                unpack_commands.append(f"cp /usr/share/ou/content/content-{idx} $HOME/{content['target']}")
            else:
                basepath, itemname = os.path.split(content["source"])
                if basepath:
                    subprocess.run(
                        [
                            "tar",
                            "-jcf",
                            os.path.join("ou-builder-build", "content", "archives", f"content-{idx}.tar.bz2"),
                            "--directory",
                            basepath,
                            itemname,
                        ]
                    )
                else:
                    subprocess.run(
                        [
                            "tar",
                            "-jcf",
                            os.path.join("ou-builder-build", "content", "archives", f"content-{idx}.tar.bz2"),
                            itemname,
                        ]
                    )
                unpack_commands.append(f"mkdir -p $HOME/{content['target']}")
                if content["overwrite"] == "always":
                    unpack_commands.append(
                        f"tar -jxf /usr/share/ou/content/content-{idx}.tar.bz2 --directory $HOME/{content['target']} --strip-components=1"
                    )
                elif content["overwrite"] == "never":
                    unpack_commands.append(
                        f"tar -jxf /usr/share/ou/content/content-{idx}.tar.bz2 --skip-old-files --directory $HOME/{content['target']} --strip-components=1"
                    )
        progress.update(content_task, advance=1)
    if len(unpack_commands) > 0:
        with open(os.path.join("ou-builder-build", "content", "020-distributing-files"), "w") as out_f:
            unpack_commmand_text = "\n".join(unpack_commands)
            out_f.write(
                f"""#!/bin/bash
{unpack_commmand_text}"""
            )
        state.update(
            {
                "output_blocks": {
                    "deploy": [
                        {
                            "block": docker_copy_cmd(
                                os.path.join("ou-builder-build", "content", "archives"), "/usr/share/ou/content"
                            ),
                            "weight": 182,
                        },
                        {
                            "block": docker_copy_cmd(
                                os.path.join("ou-builder-build", "content", "020-distributing-files"),
                                "/usr/share/ou/startup.d/020-distributing-files",
                            ),
                            "weight": 183,
                        },
                    ]
                }
            }
        )
    progress.update(content_task, visible=False)



"""CLI interface for building the container."""
import os
import shutil

from rich.progress import Progress

from ou_container_builder import core, packs
from ou_container_builder.cli.base import app
from ou_container_builder.settings import load_settings, Settings
from ou_container_builder.state import State


@app.command()
def generate():
    """Generate the Dockerfile."""
    with Progress() as progress:
        main_task = progress.add_task("Processing...", total=7)
        if os.path.exists("ou-builder-build"):
            shutil.rmtree("ou-builder-build")
        if os.path.exists("Dockerfile"):
            os.unlink("Dockerfile")
        os.makedirs("ou-builder-build")
        progress.update(main_task, advance=1)
        user_settings = load_settings()
        state = State()
        core.init(state, progress)
        packs.init(state, progress)
        progress.update(main_task, advance=1)
        progress.update(main_task, advance=1)
        state.update(user_settings)
        validated_settings = Settings(**state).model_dump()
        state = State()
        state.update(validated_settings)
        progress.update(main_task, advance=1)
        core.generate(state, progress)
        progress.update(main_task, advance=1)
        packs.generate(state, progress)
        progress.update(main_task, advance=1)
        with open("Dockerfile", "w") as out_f:
            blocks = state["output_blocks"]["build"]
            blocks.sort(key=lambda b: b["weight"])
            for block in blocks:
                out_f.write(block["block"])
                if block["block"].endswith("\n"):
                    out_f.write("\n")
                else:
                    out_f.write("\n\n")
            blocks = state["output_blocks"]["deploy"]
            blocks.sort(key=lambda b: b["weight"])
            for block in blocks:
                out_f.write(block["block"])
                if block["block"].endswith("\n"):
                    out_f.write("\n")
                else:
                    out_f.write("\n\n")
        progress.update(main_task, advance=1)



"""Functionality to install XFCE4 virtual desktop."""
import os
import shutil

from importlib.resources import open_text, open_binary
from pydantic import BaseModel
from rich.progress import Progress

from ou_container_builder.state import State
from ou_container_builder.util import docker_run_cmd, docker_copy_cmd

name = "xfce4"


class Options(BaseModel):
    """Options for the XFCE4 pack."""

    pass


def init(state: State) -> None:
    """Initialise packs.jupyterlab."""
    if os.path.exists(os.path.join("ou-builder-build", "xfce4", "xfce4", "xfconf", "xfce-perchannel-xml")):
        shutil.rmtree(os.path.join("ou-builder-build", "xfce4", "xfce4", "xfconf", "xfce-perchannel-xml"))
    os.makedirs(os.path.join("ou-builder-build", "xfce4", "xfce4", "xfconf", "xfce-perchannel-xml"), exist_ok=True)
    with open(
        os.path.join("ou-builder-build", "xfce4", "xfce4", "xfconf", "xfce-perchannel-xml", "xfce4-desktop.xml"), "w"
    ) as out_f:
        with open_text("ou_container_builder.packs.xfce4", "xfce4-desktop.xml") as in_f:
            out_f.write(in_f.read())
    with open(
        os.path.join("ou-builder-build", "xfce4", "xfce4", "xfconf", "xfce-perchannel-xml", "xfwm4.xml"), "w"
    ) as out_f:
        with open_text("ou_container_builder.packs.xfce4", "xfwm4.xml") as in_f:
            out_f.write(in_f.read())
    with open(os.path.join("ou-builder-build", "xfce4", "desktop.jpg"), "wb") as out_f:
        with open_binary("ou_container_builder.packs.xfce4", "desktop.jpg") as in_f:
            out_f.write(in_f.read())
    state.update(
        {
            "packages": {
                "apt": {
                    "deploy": [
                        "dbus-x11",
                        "libgl1-mesa-glx",
                        "xorg",
                        "xfce4",
                        "xfce4-panel",
                        "xfce4-session",
                        "xfce4-settings",
                        "at-spi2-core",
                        "tigervnc-standalone-server",
                        "viewnior",
                        "firefox-esr",
                    ]
                },
                "pip": {"system": ["jupyter-remote-desktop-proxy", "git+https://github.com/novnc/websockify.git"]},
            },
            "content": [
                {
                    "source": os.path.join("ou-builder-build", "xfce4", "xfce4"),
                    "target": ".config/xfce4",
                    "overwrite": "always",
                }
            ],
        }
    )


def generate(state: State, progress: Progress) -> None:
    """Generate the files to distribute and the Docker blocks."""
    state.update(
        {
            "output_blocks": {
                "build": [
                    {
                        "block": docker_run_cmd(
                            [
                                "git clone https://github.com/novnc/websockify.git /tmp/websockify",
                                "cd /tmp/websockify",
                                "make",
                            ]
                        ),
                        "weight": 1021,
                    }
                ],
                "deploy": [
                    {
                        "block": docker_run_cmd(
                            ["ln -s /var/lib/ou/python/system/bin/websockify /usr/local/bin/websockify"]
                        ),
                        "weight": 1021,
                    },
                    {
                        "block": docker_copy_cmd(
                            "/tmp/websockify/rebind.so", "/usr/local/bin/rebind.so", from_stage="base"
                        ),
                        "weight": 1022,
                    },
                    {
                        "block": docker_copy_cmd(
                            "ou-builder-build/xfce4/desktop.jpg", "/usr/share/backgrounds/xfce/open-university.jpg"
                        ),
                        "weight": 1023,
                    },
                ],
            }
        }
    )



"""Test the docker args settings validation."""
from ou_container_builder.settings import Arg


def test_valid_docker_args_settings():
    """Test that a valid docker args configuration passes."""
    Arg(name="TARGETPLATFORM")
    Arg(name="TARGETPLATFORM", value="linux/amd64")



"""Tests for the state management functionality."""
from unittest.mock import Mock

from ou_container_builder.state import merge_settings, State


def test_basic_merge() -> None:
    """Test that a basic merge works."""
    result = merge_settings({}, {"a": "value"})
    assert result["a"] == "value"


def test_basic_merge_with_existing_data() -> None:
    """Test that a basic merge works with existing data."""
    result = merge_settings({"b": "another value"}, {"a": "value"})
    assert result["a"] == "value"
    assert result["b"] == "another value"


def test_basic_merge_overwrite() -> None:
    """Test that a basic merge works with overwriting simple values."""
    result = merge_settings({"a": "old value"}, {"a": "new value"})
    assert result["a"] == "new value"


def test_merge_list() -> None:
    """Test that list items are merged together."""
    result = merge_settings({"a": ["one", "two"]}, {"a": ["three", "four"]})
    assert result["a"] == ["one", "two", "three", "four"]


def test_merge_nested() -> None:
    """Test that merging nested objects works."""
    result = merge_settings(
        {"a": {"a1": ["one", "two"], "a2": "old value", "a3": "keep this"}},
        {"a": {"a1": ["three", "four"], "a2": "new value"}, "b": {"b1": "something new"}},
    )
    assert result["a"]["a1"] == ["one", "two", "three", "four"]
    assert result["a"]["a2"] == "new value"
    assert result["a"]["a3"] == "keep this"
    assert result["b"]["b1"] == "something new"


def test_state_update() -> None:
    """Test that updating the state works."""
    state = State()
    state.update({"a": {"a1": ["one", "two"], "a2": "old value", "a3": "keep this"}})
    assert state["a"]["a1"] == ["one", "two"]
    assert state["a"]["a2"] == "old value"
    assert state["a"]["a3"] == "keep this"
    state.update({"a": {"a1": ["three", "four"], "a2": "new value"}, "b": {"b1": "something new"}})
    assert state["a"]["a1"] == ["one", "two", "three", "four"]
    assert state["a"]["a2"] == "new value"
    assert state["a"]["a3"] == "keep this"
    assert state["b"]["b1"] == "something new"


def test_state_basic_listener() -> None:
    """Test that a basic listener is called."""
    listener = Mock()
    state = State()
    state.add_listener("a", listener)
    state.update({"a": "new value"})
    assert listener.called


def test_multiple_listeners() -> None:
    """Test that multiple listeners are called."""
    listener1 = Mock()
    listener2 = Mock()
    state = State()
    state.add_listener("a", listener1)
    state.add_listener("a", listener2)
    state.update({"a": "new value"})
    assert listener1.called
    assert listener2.called


def test_state_length() -> None:
    """Test that the state reports the correct length."""
    state = State()
    assert len(state) == 0
    state.update({"a": "old value"})
    assert len(state) == 1
    state.update({"a": "new value"})
    assert len(state) == 1
    state.update({"b": "new value"})
    assert len(state) == 2


def test_state_iterate() -> None:
    """Test that iterating over the top-level state keys works."""
    state = State()
    for _ in state:
        assert False
    state.update({"a": "value"})
    for value in state:
        assert value == "a"



"""Test the generate and clean functions."""
import os

from typer import Typer
from typer.testing import CliRunner


def test_generate_and_clean(runner: CliRunner, app: Typer):
    """Test that generating and cleaning succeeds."""
    cwd = os.getcwd()
    try:
        os.chdir(os.path.join("tests", "demos", "jupyterlab", "v4"))
        result = runner.invoke(app, ["clean"])
        assert not os.path.exists("Dockerfile")
        assert not os.path.exists("ou-builder-build")
        result = runner.invoke(app, ["clean"])
        assert result.exit_code == 0
        assert not os.path.exists("Dockerfile")
        assert not os.path.exists("ou-builder-build")
        result = runner.invoke(app, ["generate"])
        assert result.exit_code == 0
        assert os.path.exists("Dockerfile")
        assert os.path.exists("ou-builder-build")
        result = runner.invoke(app, ["generate"])
        assert result.exit_code == 0
        assert os.path.exists("Dockerfile")
        assert os.path.exists("ou-builder-build")
        result = runner.invoke(app, ["clean"])
        assert result.exit_code == 0
        assert not os.path.exists("Dockerfile")
        assert not os.path.exists("ou-builder-build")
    finally:
        os.chdir(cwd)
